/*
    PWM and analogRead for the Raspberry Pi Pico RP2040
    Copyright (c) 2021 Earle F. Philhower, III <earlephilhower@yahoo.com>
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <Arduino.h>
#include <CoreMutex.h>
#include <hardware/gpio.h>
#include <hardware/pwm.h>
#include <hardware/clocks.h>
#include <hardware/pll.h>
#include <hardware/adc.h>

#define PWM_CHANNELS 32
//casp modified pwm
static uint32_t analogScale[PWM_CHANNELS];// = 255;
static uint32_t analogFreq[PWM_CHANNELS];// = 1000;
static bool pwmInitted[PWM_CHANNELS];// = false;
static bool adcInitted = false;
static uint32_t analogWritePseudoScale[PWM_CHANNELS];// = 1;
static uint32_t analogWriteSlowScale[PWM_CHANNELS];// = 1;
static int g_pwm_prev_val[32];

auto_init_mutex(_dacMutex);

extern "C" void analogWriteFreq(uint32_t freq, int pin)
{
    if(freq == analogFreq[pin])
        return;
    if(freq < 1) freq = 1;//casp changed from 100
    //else if(freq > 60000) freq = 60000;/casp disabled
    analogFreq[pin] = freq;
    pwmInitted[pin] = false;
    DEBUGCORE("Adjusting analogWriteFreq for pin=%s, to=%d\n", pin, freq);
}

extern "C" void analogWriteRange(uint32_t range, int pin)
{
    if(range == analogScale[pin])
        return;
    if(range < 1) range = 1;//casp changed from 16 to 1
    else if(range > 65535) range = 65535;
    analogScale[pin] = range;
    pwmInitted[pin] = false;
    DEBUGCORE("Adjusting analogWriteRange for pin=%s, to=%d\n", pin, range);
}

//casp disabled
//extern "C" void analogWriteResolution(int res) {
//    if ((res >= 4) && (res <= 16)) {
//        analogWriteRange((1 << res) - 1);
//    } else {
//        DEBUGCORE("ERROR: analogWriteResolution out of range (%d)\n", res);
//    }
//}

extern "C" void analogWrite_init(int res)//casp added
{
    for(int i=0; i<PWM_CHANNELS; ++i)
    {
        analogScale[i] = 255;
        analogFreq[i] = 1000;
        //
        analogWriteRange((1 << res) - 1, i);
        analogWriteFreq(clock_get_hz(clk_sys)/(1 << res), i);//casp changed from 19200000 to clock_get_hz(clk_sys)
        //
        pwmInitted[i] = false;
    }
}

extern "C" void analogWrite(pin_size_t pin, int val) {
    CoreMutex m(&_dacMutex);

    if ((pin > 29) || !m) {
        DEBUGCORE("ERROR: Illegal analogWrite pin (%d)\n", pin);
        return;
    }
#ifdef ARDUINO_NANO_RP2040_CONNECT
    if(pin > 24)
    {
        analogWrite_nina((NinaPin)pin, val);
        return;
    }
#endif
    if(pwmInitted[pin] == false)
    {
        // For low frequencies, we need to scale the output max value up to achieve lower periods
        analogWritePseudoScale[pin] = 1;
        while (((clock_get_hz(clk_sys) / (float)(analogScale[pin] * analogFreq[pin])) > 255.0) && (analogScale[pin] < 32678))
        {
            analogWritePseudoScale[pin]++;
            analogScale[pin] *= 2;
            DEBUGCORE("Adjusting analogWrite values PS=%d, scale=%d\n", analogWritePseudoScale[pin], analogScale[pin]);
        }
        // For high frequencies, we need to scale the output max value down to actually hit the frequency target
        analogWriteSlowScale[pin] = 1;
        while (((clock_get_hz(clk_sys) / (float)(analogScale[pin] * analogFreq[pin])) < 2.0) && (analogScale[pin] > 32))
        {
            analogWriteSlowScale[pin]++;
            analogScale[pin] /= 2;
            DEBUGCORE("Adjusting analogWrite values SS=%d, scale=%d\n", analogWriteSlowScale[pin], analogScale[pin]);
        }

        pwm_config c = pwm_get_default_config();
        pwm_config_set_clkdiv(&c, clock_get_hz(clk_sys) / (float)(analogScale[pin] * analogFreq[pin]));
        pwm_config_set_wrap(&c, analogScale[pin]);
        //casp disbled
        //for (int i = 0; i < 30; i++)
        //    pwm_init(pwm_gpio_to_slice_num(i), &c, true);
        pwm_init(pwm_gpio_to_slice_num(pin), &c, true);
        gpio_set_function(pin, GPIO_FUNC_PWM);
        pwmInitted[pin] = true;
        //set prev val
        if(pin%2)
        {
            if(pwmInitted[pin-1])
                pwm_set_gpio_level(pin-1, g_pwm_prev_val[pin-1]);
        }
        else
        {
            if(pwmInitted[pin+1])
                pwm_set_gpio_level(pin+1, g_pwm_prev_val[pin+1]);
        }
    }
    val <<= analogWritePseudoScale[pin];
    val >>= analogWriteSlowScale[pin];

    if(val < 0)
        val = 0;
    else if((uint32_t)val > analogScale[pin])
        val = analogScale[pin];
    g_pwm_prev_val[pin] = val;
    pwm_set_gpio_level(pin, val);
}

auto_init_mutex(_adcMutex);

extern "C" int analogRead(pin_size_t pin) {
    CoreMutex m(&_adcMutex);
#ifdef ARDUINO_NANO_RP2040_CONNECT
    if(pin >= NinaPin::A4)
        return analogRead_nina((NinaPin)pin);
#endif
    if ((pin < A0) || (pin > A3) || !m) {
        DEBUGCORE("ERROR: Illegal analogRead pin (%d)\n", pin);
        return 0;
    }
    if (!adcInitted) {
        adc_init();
    }
    adc_gpio_init(pin);
    adc_select_input(pin - A0);
    return adc_read();
}

extern "C" float analogReadTemp() {
    CoreMutex m(&_adcMutex);

    if (!m) {
        return 0.0f; // Deadlock
    }
    if (!adcInitted) {
        adc_init();
    }
    adc_set_temp_sensor_enabled(true);
    delay(1); // Allow things to settle.  Without this, readings can be erratic
    adc_select_input(4); // Temperature sensor
    int v = adc_read();
    adc_set_temp_sensor_enabled(false);
    float t = 27.0f - ((v * 3.3f / 4096.0f) - 0.706f) / 0.001721f; // From the datasheet
    return t;
}
